"""_fetchrepodata: Map yum/dnf repo metadata to local lookup caches"""
import gzip
import json
import logging
import os
import re
from collections import defaultdict

import click
import requests

import gi
gi.require_version('Modulemd', '1.0')
from gi.repository import Modulemd

from attr import attributes, attrib
from lxml import etree
from requests_toolbelt.downloadutils.tee import tee_to_file
from urllib.parse import urljoin

from .config import config

XDG_CACHE_HOME = os.environ.get("XDG_CACHE_HOME") or os.path.expanduser("~/.cache")
CACHEDIR = os.path.join(XDG_CACHE_HOME, "fedmod")

log = logging.getLogger(__name__)

class MissingMetadata(Exception):
    """Reports failure to find the local metadata cache"""

@attributes
class RepoPaths:
    remote_repo_url = attrib(type=str)
    remote_metadata_url = attrib(type=str)
    local_cache_path = attrib(type=str)
    local_metadata_path = attrib(type=str)

@attributes
class RepoPathsPair:
    arch = attrib(type=RepoPaths)
    src = attrib(type=RepoPaths)

def _define_repo(remote_repo_template, subst_dict, local_cache_name,
                 arch=None):

    # expand placeholders in URLs
    remote_repo_url = remote_repo_template
    for subst_key, subst_value in subst_dict.items():
        remote_repo_url = re.sub(fr"\${subst_key}", subst_value,
                                 remote_repo_url)

    local_arch_path = arch or 'source'

    remote_metadata_url = os.path.join(remote_repo_url, "repodata/")
    local_cache_path = os.path.join(CACHEDIR, "repos", local_cache_name, local_arch_path)
    local_metadata_path = os.path.join(local_cache_path, "repodata/")
    return RepoPaths(remote_repo_url, remote_metadata_url,
                     local_cache_path, local_metadata_path)

def _define_repos(arch_remote_prefix, src_remote_prefix, subst_dict, local_cache_name, arch):
    return RepoPathsPair(arch=_define_repo(arch_remote_prefix, subst_dict, local_cache_name, arch),
                         src=_define_repo(src_remote_prefix, subst_dict, local_cache_name))


class DistroPaths(object):
    def __init__(self, release_name, arch):
        self.release_name = release_name
        self.release = config.releases[release_name]
        self.arch = arch

        subst_dict = {
            'basearch': arch or 'source',
        }

        dataset_regex = self.release.get('dataset-regex')
        if dataset_regex:
            match = re.match(dataset_regex, release_name)

            if not match:
                raise ValueError(
                    f"Couldn't parse release name {release_name!r} with"
                    f" dataset-regex {dataset_regex!r}")

            subst_dict.update(match.groupdict())

        self.repo_paths_by_name = {
            repo_name: _define_repos(repo_def['arch']['baseurl'],
                                     repo_def['source']['baseurl'],
                                     subst_dict,
                                     release_name + "--" + repo_name, arch)
            for repo_name, repo_def in
            config.releases[release_name]['repositories'].items()
        }


def _get_default_dataset():
    return config.get('options', {}).get('dataset')


def parse_dataset_name(dataset_name):
    """Parse dataset_name as distro_release[-arch] and validate the result"""

    if dataset_name is None:
        dataset_name = _get_default_dataset()

    parts = dataset_name.split("-")
    if len(parts) == 1:
        release_name, arch = parts[0], None
    elif len(parts) == 2:
        release_name, arch = parts[0], parts[1]
    else:
        raise ValueError("dataset name should have the form RELEASE[-ARCH]")

    try:
        dataset_config = config.releases[release_name]
    except KeyError:
        raise ValueError("Unknown release name: {}. Known releases: {}"
                         .format(release_name, ", ".join(
                             sorted(config['datasets']['releases']))))
    if not arch:
        arch = dataset_config.get('default-architecture', 'x86_64')

    if arch not in dataset_config['architectures']:
        raise ValueError("Unknown architecture: {}. Known releases: {}".format(
                arch, ", ".join(sorted(dataset_config['architectures']))))

    return release_name, arch

def _display_dataset_name(dataset_name):
    release_name, arch = parse_dataset_name(dataset_name)
    if (release_name, arch) == parse_dataset_name(_get_default_dataset()):
        return None
    elif arch == config['datasets']['releases']['default-architecture']:
        return release_name
    else:
        return dataset_name

def _get_distro_paths(dataset_name):
    release_name, arch = parse_dataset_name(dataset_name)
    return DistroPaths(release_name, arch)

_MODULE_FORWARD_LOOKUP_CACHE = "module-contents"
_SRPM_REVERSE_LOOKUP_CACHE = "srpm-to-module"
_RPM_REVERSE_LOOKUP_CACHE = "rpm-to-module"

_ALL_CACHES = [_MODULE_FORWARD_LOOKUP_CACHE,
               _SRPM_REVERSE_LOOKUP_CACHE,
               _RPM_REVERSE_LOOKUP_CACHE]

METADATA_SECTIONS = ("filelists", "primary", "modules")

_REPOMD_XML_NAMESPACE = {"rpm": "http://linux.duke.edu/metadata/repo"}
def _read_repomd_location(repomd_xml, section):
    location = repomd_xml.find(f"rpm:data[@type='{section}']/rpm:location", _REPOMD_XML_NAMESPACE)
    if location is not None:
        return location.attrib["href"]
    return None

def _download_one_file(remote_url, filename):
    if os.path.exists(filename) and not filename.endswith((".xml", ".yaml")):
        print(f"  Skipping download; {filename} already exists")
        return
    response = requests.get(remote_url, stream=True)
    try:
        print(f"  Downloading {remote_url}")
        chunksize = 65536
        try:
            expected_chunks = int(
                response.headers['content-length']) / chunksize
        except (KeyError, ValueError):
            expected_chunks = None
        downloader = tee_to_file(response, filename=filename, chunksize=chunksize)
        show_progress = click.progressbar(downloader, length=expected_chunks)
        with show_progress:
            for chunk in show_progress:
                pass
    finally:
        response.close()
    print(f"  Added {filename} to cache")

def _download_metadata_files(repo_paths):
    local_path = repo_paths.local_cache_path
    local_metadata_path = repo_paths.local_metadata_path
    os.makedirs(local_metadata_path, exist_ok=True)
    repomd_url = urljoin(repo_paths.remote_metadata_url, "repomd.xml")
    print(f"Remote metadata: {repomd_url}")
    response = requests.get(repomd_url)
    response.raise_for_status()
    repomd_filename = os.path.join(local_metadata_path, "repomd.xml")
    with open(repomd_filename, "wb") as f:
        f.write(response.content)
    print(f"  Cached metadata in {repomd_filename}")
    repomd_xml = etree.parse(repomd_filename)
    files_to_fetch = set()
    for section in METADATA_SECTIONS:
        relative_href = _read_repomd_location(repomd_xml, section)
        if relative_href is not None:
            files_to_fetch.add(relative_href)
    predownload = set(os.listdir(local_path))
    for relative_href in files_to_fetch:
        absolute_href = urljoin(repo_paths.remote_repo_url, relative_href)
        filename = os.path.join(local_path, relative_href)
        # This could be parallelised with concurrent.futures, but
        # probably not worth it (it makes the progress bars trickier)
        _download_one_file(absolute_href, filename)
    postdownload = set(os.listdir(local_path))
    # Prune any old metadata files automatically
    if len(postdownload) >= (len(predownload) + len(METADATA_SECTIONS)):
        # TODO: Actually prune old metadata files
        pass

def _cache_fname(paths, cache_name):
    return os.path.join(CACHEDIR, f"{paths.release_name}-{cache_name}-{paths.arch}-cache.json")

def _write_cache(paths, cache_name, data):
    """Write the given data to the nominated cache file"""
    cache_fname = _cache_fname(paths, cache_name)
    with open(cache_fname, "w") as cache_file:
        json.dump(data, cache_file)
    print(f"  Added {cache_fname} to cache")

def _read_cache(paths, cache_name):
    """Read the parsed data from the nominated cache file"""
    cache_fname = _cache_fname(paths, cache_name)
    with open(cache_fname, "r") as cache_file:
        return json.load(cache_file)

def _read_modules(repo_paths):
    metadata_dir = os.path.join(repo_paths.local_cache_path)
    repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")
    repomd_xml = etree.parse(repomd_fname)
    repo_relative_modulemd = _read_repomd_location(repomd_xml, "modules")

    if not repo_relative_modulemd:
        # repository doesn't contain modules
        return

    repo_modulemd_fname = os.path.join(metadata_dir, repo_relative_modulemd)

    with gzip.open(repo_modulemd_fname, "rt") as modules_yaml_gz:
        modules_yaml = modules_yaml_gz.read()

    objects = Modulemd.objects_from_string(modules_yaml)
    return (o for o in objects if isinstance(o,  Modulemd.Module))

def _merge_modules(module_sets):
    modules = dict()

    for module_set in module_sets:
        for m in module_set:
            old = modules.get((m.props.name, m.props.stream))
            if not old or m.props.version > old.props.version:
                modules[(m.props.name, m.props.stream)] = m

    return modules.values()

def _write_lookup_caches(paths):
    module_sets = [modules_read
                   for modules_read in (
                       _read_modules(repopaths.arch)
                       for repopaths in paths.repo_paths_by_name.values())
                   if modules_read]

    module_forward_lookup = {}
    srpm_reverse_lookup = defaultdict(list)
    rpm_reverse_lookup = defaultdict(list)

    for module in _merge_modules(module_sets):
        module_name = module.props.name
        artifacts = module.props.rpm_artifacts.get()

        module_forward_lookup[module_name] = list(set(artifacts))
        components = module.props.components_rpm
        for srpmname in components:
            srpm_reverse_lookup[srpmname].append(module_name)
        for rpmname in artifacts:
            rpmprefix = rpmname.split(":", 1)[0].rsplit("-", 1)[0]
            rpm_reverse_lookup[rpmprefix].append(module_name)

    # Cache the lookup tables as local JSON files
    print("Caching lookup tables")
    _write_cache(paths, _MODULE_FORWARD_LOOKUP_CACHE, module_forward_lookup)
    _write_cache(paths, _SRPM_REVERSE_LOOKUP_CACHE, srpm_reverse_lookup)
    _write_cache(paths, _RPM_REVERSE_LOOKUP_CACHE, rpm_reverse_lookup)


def download_repo_metadata(dataset_name):
    """Downloads the latest repo metadata"""

    paths = _get_distro_paths(dataset_name)
    for repo_pair in paths.repo_paths_by_name.values():
        for repo_definition in (repo_pair.arch, repo_pair.src):
            _download_metadata_files(repo_definition)
    _write_lookup_caches(paths)

@attributes
class LocalMetadataCache:
    dataset_name = attrib(type=str)
    cache_dir = attrib(type=str)
    srpm_to_modules = attrib(type=dict)
    rpm_to_modules = attrib(type=dict)
    module_to_packages = attrib(type=dict)
    repo_cache_paths = attrib(type=dict)


def load_cached_repodata(dataset_name):
    paths = _get_distro_paths(dataset_name)

    display_name = _display_dataset_name(dataset_name)
    arg = f" --dataset={display_name}" if display_name else ""

    # Check whether or not fetch-metadata has been run at all
    for repo_name, repo_paths in paths.repo_paths_by_name.items():
        metadata_dir = os.path.join(repo_paths.arch.local_cache_path)
        repomd_fname = os.path.join(metadata_dir, "repodata", "repomd.xml")

        if not os.path.exists(repomd_fname):
            raise MissingMetadata(f"{repomd_fname!r} does not exist. Run "
                                  f"`fedmod{arg} fetch-metadata`.")

    # Check whether or not fetch-metadata actually finished
    for cache in _ALL_CACHES:
        cache_fname = _cache_fname(paths, cache)
        if not os.path.exists(cache_fname):
            msg = (f"{cache_fname!r} does not exist. "
                   f"Try running `fedmod{arg} fetch-metadata` again.")
            raise MissingMetadata(msg)

    # Load the metadata
    return LocalMetadataCache(
        dataset_name=dataset_name,
        cache_dir=CACHEDIR,
        srpm_to_modules=_read_cache(paths, _SRPM_REVERSE_LOOKUP_CACHE),
        rpm_to_modules=_read_cache(paths, _RPM_REVERSE_LOOKUP_CACHE),
        module_to_packages=_read_cache(paths, _MODULE_FORWARD_LOOKUP_CACHE),
        repo_cache_paths={
            n: (c.arch.local_cache_path, c.src.local_cache_path)
            for n, c in paths.repo_paths_by_name.items()
        }
    )
