import sys
import click
import click_completion
import logging

from .config import config
from .flatpak_generator import FlatpakGenerator, do_flatpak_report
from .module_generator import ModuleGenerator
from .module_repoquery import ModuleRepoquery

from . import modulemd_linter as mmdl
from . import _depchase, _repodata, _fetchrepodata

# fedmod uses click for argument parsing, but currently does its own
# standardised exception handling. This also requires handling click's standard
# exception types as described in http://click.pocoo.org/5/exceptions/

def run():
    click_completion.init()
    try:
        rc = _cli_commands.main(sys.argv[1:], standalone_mode=False,
                                default_map=config.get('options', {}))
    except (KeyboardInterrupt, click.Abort):
        print('\nInterrupted by user')
        rc = 1
    except click.ClickException as e:
        e.show()
        rc = e.exit_code
    except _fetchrepodata.MissingMetadata as e:
        print(e, file=sys.stderr)
        rc = 2
    except Exception as e:
        logging.exception(f"Unexpected exception: {e!r}")
        rc = 3
    sys.exit(rc)


@click.group()
@click.option("--verbose", "-v", is_flag=True, default=False,
              help="Show additional info messages")
@click.option("--dataset", "-s", metavar="RELEASE[:ARCH]",
              help="Set the release to use for package and module data",
              type=click.Choice(config.datasets))
def _cli_commands(verbose, dataset):
    """Utilities for working with modular Fedora RPM repositories"""
    if verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARN)

    if dataset:
        try:
            _repodata.set_dataset_name(dataset)
        except ValueError as e:
            raise click.ClickException(e) from e


# Fetching metadata
@_cli_commands.command('fetch-metadata')
def fetch_metadata():
    """Fetch and cache required module and RPM metadata"""
    _fetchrepodata.download_repo_metadata(_repodata.dataset_name)


# modulemd generation
@_cli_commands.command()
@click.option("--output", "-o", metavar="FILE", help="Write to FILE instead of stdout.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def rpm2module(pkgs, output):
    """Generate a draft modulemd from an RPM (or list of RPMs)"""
    mg = ModuleGenerator(pkgs)
    mg.run(output)

# modulemd generation for Flatpaks
@_cli_commands.command()
@click.option("--output-modulemd", metavar="FILE", help="Write modulemd to FILE instead of <pkg>.yaml.")
@click.option("--output-containerspec", metavar="FILE", help="Write container specification to FILE instead of container.yaml")
@click.option("--force", "-f", is_flag=True, help="Overwriting existing output files")
@click.argument("pkg", metavar='PKG', required=True)
def rpm2flatpak(pkg, output_modulemd, output_containerspec, force):
    """Generate modulemd from an RPM"""
    fg = FlatpakGenerator(pkg)
    fg.run(output_modulemd, output_containerspec, force=force)

@_cli_commands.command('flatpak-report')
@click.option("--quiet", "-q", is_flag=True,
              help="Don't print progress messages to stderr.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def flatpak_report(pkgs, quiet):
    do_flatpak_report(pkgs, quiet=quiet)

# Checking availability of dependencies through module streams
@_cli_commands.command('resolve-deps')
@click.option("--module-dependency", "-m", multiple=True, metavar="MODULE",
            help="Module to be used as a dependency. Can be given multiple times.")
@click.option("--json", is_flag=True, default=False,
              help="Output dependencies in JSON format with extra information.")
@click.argument("pkgs", metavar='PKGS', nargs=-1, required=True)
def resolve_deps(pkgs, module_dependency, json):
    """Report dependencies of the given RPM (or list of RPMs)"""
    rq = ModuleRepoquery()
    rq.list_pkg_deps(pkgs, module_dependency, json)


# Listing modules
@_cli_commands.command('list-modules')
def list_modules():
    """List the available modules"""
    rq = ModuleRepoquery()
    rq.list_modules()


# Looking for RPM duplication across modules
@_cli_commands.command('list-rpms')
@click.option("--duplicate-only", is_flag=True, default=False,
            help="Only list packages that are in more than one module.")
@click.option("--list-modules", is_flag=True, help="List modules for every package.")
def list_rpms(duplicate_only, list_modules):
    """List RPMs available through modules"""
    rq = ModuleRepoquery()
    rq.list_modularized_pkgs(
        duplicate_only=duplicate_only,
        list_modules=list_modules,
    )

# Checking module contents
# (ncoghlan): Perhaps `module-rpms` would be more consistent?
@_cli_commands.command('module-packages')
@click.option("--full-nevra", is_flag=True, default=False,
            help="Print the full NEVRA instead of only name.")
@click.argument("module")
def module_rpms(module, full_nevra):
    """Lists RPMS in a given module"""
    rq = ModuleRepoquery()
    rq.list_rpms_in_module(module, full_nevra=full_nevra)


# Checking availability of specific packages
# (ncoghlan): Perhaps `where-is-rpm` would be more consistent?
@_cli_commands.command('where-is-package')
@click.argument("pkg")
def which_module(pkg):
    """Reports which module (if any) provides the given RPM"""
    rq = ModuleRepoquery()
    rq.list_modules_for_rpm(pkg)


# Checking what to build to satisfy particular binary dependencies
@_cli_commands.command('srpm-of-rpm')
@click.argument("pkg")
def which_srpm(pkg):
    """Reports which SRPM produces the given RPM"""
    rq = ModuleRepoquery()
    rq.get_srpm_for_rpm(pkg)

# Map SRPM to the RPMs it generates. Useful info when building
# new modules, choosing the API or what to filter
@_cli_commands.command('rpms-from-srpm')
@click.argument("pkg")
def rpms_from_srpm(pkg):
    """Reports which RPMS are generated from the given SRPM"""
    rq = ModuleRepoquery()
    rq.get_rpms_for_srpm(pkg)

# Check a modulemd YAML file for validity.
@_cli_commands.command('lint')
@click.option("--min-level", metavar="MINLEVEL", default='info',
              type=click.Choice(mmdl.LintLevel.all_levels),
              help="The minimum level of checks to be performed ({}).".format(
              ", ".join(mmdl.LintLevel.all_levels)))
@click.argument("modulemd")
def lint(modulemd, min_level):
    """Validates a given modulemd YAML file"""
    linter = mmdl.ModuleMDLinter(modulemd_path=modulemd)
    return linter.lint(min_level=min_level)
